import re
import warnings

from flap.utils import repl_positions


def is_postcode(string):
    postcode_regex = r'^([A-Z]{1,2})([0-9][A-Z0-9]?) ?([0-9][A-Z]{2})$'
    if isinstance(string, str):
        return bool(re.match(postcode_regex, string))
    else:
        return False


def postcode_preproc(string, suppress_warnings=False):
    postcode_regex = r'^([A-Z]{1,2})([0-9][A-Z0-9]?) ?([0-9][A-Z]{2})$'

    if is_postcode(string):
        return re.sub(pattern=postcode_regex, repl=r'\1\2 \3', string=string)
    else:
        if not suppress_warnings:
            warnings.warn('%s is not a postcode' % string)
        return string


def address_line_preproc(address):
    if not isinstance(address, str):
        return address, {'address_preproc': address, 'proceed': False}
    elif len(address) == 0:
        return address, {'address_preproc': address, 'proceed': False}

    preproc_mapping = [
        (r"[^\w\-/\s'\.\(\)]", ' '),
        (r'  +', ' '),
        (r'--+', '-'),
        (r'//+', '/')
    ]

    n_sub = 0
    # n_abbr = 0
    # add_pt = 0
    # add_dl = 0
    # remove_region = 0

    for pattern, repl in preproc_mapping:
        address, n = re.subn(pattern=pattern, repl=repl, string=address)
        n_sub += n

    address = address.upper()

    #
    # parsed = rule_parsing(address)
    # address_rp = parsing_to_string(parsed, address)
    #
    #
    # if len(parsed['REGION']):
    #     start, end = parsed['REGION'][0]
    #     address = repl_positions(address, list(range(start, end)))
    #     remove_region += 1
    #
    # if not len(parsed['POST_TOWN']):
    #
    #     parsed = rule_parsing(address)
    #
    #     if len(parsed['DEPENDENT_LOCALITY']) and len(parsed['POSTCODE']):
    #         start_dl, end_dl = parsed['DEPENDENT_LOCALITY'][0]
    #         start_pc, end_pc = parsed['POSTCODE'][0]
    #         dlpc0 = ':'.join([address[start_dl:end_dl],
    #                           address[start_pc:end_pc].split(' ')[0]])
    #
    #         if dlpc0 in dlpc0_to_pt:
    #             address = ' '.join([address[:end_dl], dlpc0_to_pt[dlpc0],
    #                                 address[end_dl:]])
    #
    #             add_pt += 1
    #
    #     elif len(parsed['DOUBLE_DEPENDENT_LOCALITY']) and len(parsed['POSTCODE']):
    #         start_ddl, end_ddl = parsed['DOUBLE_DEPENDENT_LOCALITY'][0]
    #
    #         if address[start_ddl:end_ddl] in ddl_to_dlpc0:
    #             dlpc0 = ddl_to_dlpc0[address[start_ddl:end_ddl]]
    #             pt = dlpc0_to_pt[dlpc0]
    #
    #             address = ' '.join([address[:end_ddl],
    #                                 dlpc0.split(':')[0], pt, address[end_ddl:]])
    #
    #             add_pt += 1
    #             add_dl += 1
    #
    # address, _ = re.subn(r' +', ' ', address.replace('*', '').upper())
    #
    log = {'address_preproc': address, 'preproc_n_sub': n_sub, 'proceed': True
           # 'preproc_n_abbr': n_abbr, 'add_pt': add_pt,
           # 'add_dl': add_dl, 'remove_region': remove_region
           }

    return address, log



from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment

import numpy as np

from flap.utils import tokens_to_span_classes, tokenize_default
from flap.utils import Span


def string_to_2d_array(s):
    return np.array(list(s)).reshape(-1, 1)


def bin_dist(x, y):
    return int(x == y)


class LinearAssignmentAlignment:

    def __init__(self, seq1, seq2):
        self.seq1 = seq1

        self.seq2_index = {}

        if isinstance(seq2, str):
            self.seq2 = seq2
        elif isinstance(seq2, dict):
            tmp_start = 0
            self.seq2 = ' '.join([v for v in seq2.values() if len(v)])

            for k, v in seq2.items():

                if len(v):
                    self.seq2_index[k] = Span(tmp_start, tmp_start + len(v))
                    tmp_start += len(v) + 1
                else:
                    self.seq2_index[k] = []

        self.cost_matrix = None
        self.assignment_mapping = None
        self.diag_paths = None

    def _prepare_cost_matrix(self, diag_bonus=True, token_bonus=True):
        a1 = string_to_2d_array(self.seq1)
        a2 = string_to_2d_array(self.seq2)

        self.cost_matrix = cdist(a1, a2, metric=bin_dist)

        if diag_bonus:
            self._apply_diag_bonus()

        if token_bonus:
            self._apply_token_bonus()

    def _find_diag_paths(self):
        x_map = self.cost_matrix.copy()

        paths = []

        n_row = x_map.shape[0]
        n_col = x_map.shape[1]

        for i in range(n_row):
            for j in range(n_col):

                if x_map[i, j]:

                    start = (i, j)
                    pi, pj = start

                    try:
                        while x_map[pi, pj]:
                            x_map[pi, pj] = 0
                            pi += 1
                            pj += 1
                    except IndexError:
                        pi -= 1
                        pj -= 1
                        end = (pi, pj)
                        x_map[pi, pj] = 0
                    else:
                        pi -= 1
                        pj -= 1
                        end = (pi, pj)
                        x_map[pi, pj] = 0

                    paths.append([start, end])

        self.diag_paths = paths

    def _apply_diag_bonus(self):

        if self.diag_paths is None:
            self._find_diag_paths()

        for path in self.diag_paths:

            (i, j), (i_end, j_end) = path
            len_path_bonus = i_end - i

            while (i <= i_end) and (j <= j_end):
                self.cost_matrix[i, j] += len_path_bonus

                i += 1
                j += 1

    def _apply_token_bonus(self, tokenizer=tokenize_default):

        if self.diag_paths is None:
            self._find_diag_paths()

        i_spans = []
        j_spans = []

        for path in self.diag_paths:
            (i0, j0), (i1, j1) = path
            i_spans.append(Span(i0, i1))
            j_spans.append(Span(j0, j1))

        t1 = tokenizer(self.seq1)
        t2 = tokenizer(self.seq2)

        sp1 = tokens_to_span_classes(t1)
        sp2 = tokens_to_span_classes(t2)

        for span in sp1:
            if any([span in i_span for i_span in i_spans]):
                self.cost_matrix[span.start:span.end + 1, :] *= 2

        for span in sp2:
            if any([span in j_span for j_span in j_spans]):
                self.cost_matrix[:, span.start:span.end + 1] *= 2

    def diag_paths_to_spans(self):
        i_spans = []
        j_spans = []

        for path in self.diag_paths:
            (i0, j0), (i1, j1) = path
            i_spans.append(Span(i0, i1))
            j_spans.append(Span(j0, j1))

        return i_spans, j_spans

    def _align(self):
        if self.cost_matrix is None:
            self._prepare_cost_matrix()

        self.assignment_mapping = [(i, j) for i, j in
                                   np.array(linear_sum_assignment(self.cost_matrix, maximize=True)).transpose()
                                   if self.seq1[i] == self.seq2[j]]

    def get_result(self):

        if self.assignment_mapping is None:
            self._align()

        return NonSequentialAlignmentResult(self.assignment_mapping, self.seq1, self.seq2, self.seq2_index)


class NonSequentialAlignmentResult:

    def __init__(self, mapping, seq1, seq2, seq2_index=None):
        self.mapping = mapping
        self.seq1 = seq1
        self.seq2 = seq2
        self.seq2_index = seq2_index

        self.meta = dict(mapping=mapping,
                         n_aligned=len(mapping),
                         ft_len_x=len(seq1),
                         ft_len_y=len(seq2))

    def __str__(self):
        mapping = self.meta['mapping']

        mapping_dict_x_to_y = dict(mapping)
        mapping_dict_y_to_x = {v: k for k, v in dict(mapping).items()}

        seq1 = self.seq1
        seq2 = self.seq2

        lines = []

        rx = seq1

        ry = ''.join(
            [seq2[mapping_dict_x_to_y[i]] if i in mapping_dict_x_to_y.keys() else '_' for i in range(len(seq1))]
        )

        lines.extend(['On Seq 1:', rx, ry])

        rx = ''.join(
            [seq1[mapping_dict_y_to_x[i]] if i in mapping_dict_y_to_x.keys() else '_' for i in range(len(seq2))])
        ry = seq2

        lines.extend(['On Seq 2:', rx, ry])

        return '\n'.join(lines)

    def __repr__(self):
        type_ = type(self)
        module = type_.__module__
        qualname = type_.__qualname__

        return '\n'.join([f'<{module}.{qualname} object at {hex(id(self))}>',
                          str(self)])

    def count_insertion(self):

        mapping = self.meta['mapping']

        try:
            mapping_ends = [[min(mapping, key=lambda x: x[0])[0], max(mapping, key=lambda x: x[0])[0]],
                            [min(mapping, key=lambda x: x[1])[1], max(mapping, key=lambda x: x[1])[1]]]
        except ValueError:
            self.meta['ft_n_insertion'] = 0
            return

        if len(self.meta['mapping']):
            n_insertion = 0

            for i in range(2):
                n_insertion += mapping_ends[i][1] - mapping_ends[i][0] - len(mapping) + 1

        else:
            n_insertion = 0

        self.meta['ft_n_insertion'] = n_insertion

    def calculate_subscore(self):

        seq2_index = self.seq2_index
        mapping = self.mapping

        seq2_backtrace = {k: [] for k in seq2_index}
        seq2_perc = {}

        for i, j in mapping:

            for k, span in seq2_index.items():

                if j in span:
                    seq2_backtrace[k].append(i)
                    break

        for k in seq2_backtrace:
            if len(seq2_backtrace[k]):
                seq2_perc['ft_perc_%s' % k] = len(seq2_backtrace[k]) / len(seq2_index[k])
            else:
                seq2_perc['ft_perc_%s' % k] = 0

        self.meta.update(seq2_perc)

    def calculate_score(self):

        if 'ft_n_insertion' not in self.meta:
            self.count_insertion()

        n_aligned = self.meta['n_aligned']
        len_x = self.meta['ft_len_x']
        len_y = self.meta['ft_len_y']

        try:
            perc_x = n_aligned/len_x
            perc_y = n_aligned/len_y
            f1 = 2 * perc_x * perc_y / (perc_x + perc_y)

        except ZeroDivisionError:
            perc_x = 0
            perc_y = 0
            f1 = 0

        self.meta['ft_perc_x'] = perc_x
        self.meta['ft_perc_y'] = perc_y
        self.meta['ft_f1'] = f1

        if self.seq2_index is not None:
            self.calculate_subscore()

        self.meta['score'] = self.meta['ft_f1']

    def get_score(self):

        self.calculate_score()
        score = {k: self.meta[k] for k in self.meta if 'ft_' in k}

        return score


#
# uprn = {'CHANGE_TYPE': 'I', 'UPRN': '10091986837', 'UDPRN': '8253456',
#         'ORGANISATION_NAME': 'MANSFIELD CARE ADMINISTRATION', 'DEPARTMENT_NAME': '', 'SUB_BUILDING_NAME': '',
#         'BUILDING_NAME': '', 'BUILDING_NUMBER': '99', 'DEPENDENT_THOROUGHFARE': '', 'THOROUGHFARE': 'CRAIGHALL ROAD',
#         'DOUBLE_DEPENDENT_LOCALITY': '', 'DEPENDENT_LOCALITY': '', 'POST_TOWN': 'EDINBURGH', 'POSTCODE': 'EH6 4RD',
#         'POSTCODE_TYPE': 'S', 'DELIVERY_POINT_SUFFIX': '1T', 'WELSH_DEPENDENT_THOROUGHFARE': '',
#         'WELSH_THOROUGHFARE': '', 'WELSH_DOUBLE_DEPENDENT_LOCALITY': '', 'WELSH_DEPENDENT_LOCALITY': '',
#         'WELSH_POST_TOWN': '', 'PO_BOX_NUMBER': '', 'PROCESS_DATE': '2018-11-06',
#         'START_DATE': '2018-12-06', 'END_DATE': '', 'LAST_UPDATE_DATE': '2018-12-06',
#         'ENTRY_DATE': '2012-03-19', 'GEOMETRY': None, 'n_tenement': '1', 'type_of_micro': '10001',
#         'type_of_macro': '0100', 'number_like_0': '99', 'number_like_1': '', 'number_like_2': '', 'number_like_3': '',
#         'number_like_4': '', 'pc0': 'EH6', 'pc1': '4RD', 'pc_area': 'EH', 'pc_district': '6', 'pc_sector': '4',
#         'pc_unit_0': 'R', 'pc_unit_1': 'D'}

#
# from src.database.sql import SqlDB
#
#
# sql_db = SqlDB('/home/huayu_ssh/PycharmProjects/dres_r/db/scotland_curl')
#
# columns = ['ORGANISATION_NAME', 'DEPARTMENT_NAME', 'SUB_BUILDING_NAME', 'BUILDING_NAME', 'BUILDING_NUMBER',
# 'DEPENDENT_THOROUGHFARE', 'THOROUGHFARE', 'DOUBLE_DEPENDENT_LOCALITY', 'DEPENDENT_LOCALITY', 'POST_TOWN', 'POSTCODE']
#
#
# res = sql_db.sql_query('select * from indexed limit 1')[0]
#
# d = {k: uprn[k] for k in columns}
#
# laa = LinearAssignmentAlignment(seq1='MANSFIELD CARE ADMINISTRATION 99 CRAIGHALL ROAD EDINBURGH EH6 4RD',
#                                 seq2=d)
#
# ar = laa.get_result()


#
# laa.seq2
# laa.seq2_index
#
#
# ar = laa.get_result()
#
#
# mapping = ar.mapping
#
# seq2 = ar.seq2
# seq2_index = ar.seq2_index
#
# seq2_backtrace = {k: [] for k in seq2_index}
# seq2_perc = {}
#
#
# for i, j in mapping:
#
#     for k, span in seq2_index.items():
#
#         if j in span:
#
#             seq2_backtrace[k].append(i)
#             break
#
# for k in seq2_backtrace:
#     if len(seq2_backtrace[k]):
#         seq2_perc[k] = len(seq2_backtrace[k])/len(seq2_index[k])
#     else:
#         seq2_perc[k] = 0
#



import time
import os
from tqdm.contrib.concurrent import process_map


def _test_func(x):
    time.sleep(0.01)
    return x + 2


def calc_chunksize(n_workers, len_iterable, factor=4):
    """Calculate chunksize argument for Pool-methods.

    Resembles source-code within `multiprocessing.pool.Pool._map_async`.
    """
    chunksize, extra = divmod(len_iterable, n_workers * factor)
    if extra:
        chunksize += 1
    return chunksize


def _progress_map_naive(func, iterable):
    map_result = []

    n = len(iterable)
    n_parts = min(20, n)
    n_tile = int(n / n_parts)
    tile_counter = 0
    counter = 0
    counter_c = 0
    t0 = time.time()
    tc = time.time()
    refresh_interval = 1

    for i in iterable:

        map_result.append(func(i))

        counter += 1
        if counter % n_tile == 0:
            tile_counter += 1

        if (time.time() - tc) > refresh_interval:
            progress = '+' * tile_counter + '>' + '-' * (n_parts - tile_counter - 1)
            print('\r[%s](%s/%s) [%.4f iter/s, time elapsed = %.2f s]' % (
                progress, counter, n, (counter - counter_c) / (time.time() - tc), time.time() - t0), end='\r')
            counter_c = counter
            tc = time.time()

    tile_counter += 1
    progress = '+' * tile_counter
    print('\r[%s](%s/%s) [%.4f s, total time = %.2f s]' % (
        progress, counter, n, counter / (time.time() - t0), time.time() - t0), end='\r')

    return map_result


def _progress_map_tqdm(func, iterable):
    from tqdm import tqdm
    map_result = [func(i) for i in tqdm(iterable)]
    return map_result


def progress_map_tqdm_concurrent(func, iterable, length=None, max_workers=None, chunksize=None):

    if max_workers is None:
        max_workers = os.cpu_count()

    if chunksize is None:
        if length is not None:
            chunksize = calc_chunksize(max_workers, length)
        else:
            try:
                chunksize = calc_chunksize(max_workers, len(iterable))
            except TypeError:
                chunksize = calc_chunksize(max_workers, 5000)

    map_result = process_map(func, iterable, max_workers=max_workers,
                             chunksize=chunksize)

    return map_result


def progress_map(func, iterable):

    try:
        map_result = _progress_map_tqdm(func, iterable)
    except ModuleNotFoundError:
        map_result = _progress_map_naive(func, iterable)

    return map_result



import re
import json

from copy import deepcopy

from flap.utils import span_to_string, repl_char_from_string, span_end, flatten
from flap.database.sql import SqlDB


class RuleParserFast:

    @staticmethod
    def load_json(path):
        with open(path, 'r') as f:
            loaded = json.load(f)
        return loaded

    def __init__(self, sql_db=None):

        self.l_thoroughfare_names = None
        self.l_posttown_names = None
        self.pc1_mappings = None
        self.pc1_mappings_region = None

        self.regex_rules = None
        self.need_fuzzy_parsing = False

        if sql_db is not None:
            self.l_thoroughfare_names = self.load_json(sql_db.sub_paths['thoroughfare_patterns'])
            self.l_posttown_names = self.load_json(sql_db.sub_paths['unique_POST_TOWN'])
            self.pc1_mappings = self.load_json(sql_db.sub_paths['pc1_mappings'])
            self.pc1_mappings_region = self.load_json(sql_db.sub_paths['pc1_mappings_region'])

            self.pc1_mappings['REGION'] = self.pc1_mappings_region

            self.regex_rules = {
                'POSTCODE': [r'([A-Z]{1,2}[0-9][A-Z0-9]?)(?: *\n?)?([0-9][A-Z]{2})'],
                'POST_TOWN': [r'(?s:.*)(?:\s|,|\n)(%s)(?:\s|$|,|\n)?' % s for s in self.l_posttown_names],
                'REGION': [r'(?s:.*)(?:\s|,|\n)(%s)(?:\s|$|,|\n)?' % s
                           for s in ['FIFE', 'TAYSIDE', 'LOTHIAN', 'WEST LOTHIAN', 'EAST LOTHIAN', 'MIDLOTHIAN']],
                'THOROUGHFARE': [
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'\d+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'\d+[A-Z]+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'[A-Z]+\d+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'\d+(?:/|-)\d+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'\d+(?:/|-)[A-Z]+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'[A-Z]+(?:/|-)\d+', '|'.join(self.l_thoroughfare_names)),
                    r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % (r'[A-Z]+(?:/|-)[A-Z]+', '|'.join(self.l_thoroughfare_names)),
                    r'((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % '|'.join(self.l_thoroughfare_names),
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'\d+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'\d+[A-Z]+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'[A-Z]+\d+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'\d+(?:/|-)\d+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'\d+(?:/|-)[A-Z]+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'[A-Z]+(?:/|-)\d+',
                    r'(%s) ((?:[A-Z]+ ){0,2}[A-Z]+)(?:$|,|\n)' % r'[A-Z]+(?:/|-)[A-Z]+'
                ]
            }

    def copy(self):
        return deepcopy(self)

    def exact_parsing(self, address):

        exact_parsing_fields = ['REGION', 'THOROUGHFARE', 'DOUBLE_DEPENDENT_LOCALITY', 'DEPENDENT_LOCALITY', 'POST_TOWN']

        string = {}
        span = {}

        need_fuzzy_parsing = False
        residual_address = address

        try:

            match = list(re.finditer(self.regex_rules['POSTCODE'][0], address))[0]

            string['POSTCODE'] = ' '.join([match.group(1), match.group(2)])
            string['PC1'] = match.group(1)
            string['PC2'] = match.group(2)

            span['POSTCODE'] = match.span(0)

            for field in exact_parsing_fields:

                if string['PC1'] in self.pc1_mappings[field]:
                    l_patterns = self.pc1_mappings[field][string['PC1']]
                    l_regex = [r'(?s:.*)(%s)' % s for s in l_patterns]

                    match_list = []

                    for regex in l_regex:
                        match = re.search(pattern=regex, string=address)
                        if match is not None:
                            match_list.append(match)

                    if len(match_list):
                        match = max(match_list, key=lambda x: len(x.group(1)))
                        string[field] = match.group(1)
                        span[field] = match.span(1)

            for key in ['THOROUGHFARE', 'POST_TOWN']:
                if key not in string:
                    need_fuzzy_parsing = True

            for sp in span.values():
                residual_address = repl_char_from_string(residual_address, sp)

        except IndexError:
            need_fuzzy_parsing = True

        string['residual_address'] = residual_address

        return string, need_fuzzy_parsing

    def fuzzy_parsing(self, address):

        span = {}
        string = {}

        residual_address = address

        for key, l_regex in self.regex_rules.items():

            match_type = 0

            if key == 'THOROUGHFARE':

                l_match_results = []

                for regex in l_regex:

                    for i in range(2):
                        match = re.search(pattern=regex, string=residual_address)

                        if match:

                            l_match_results.append(
                                {'match': match,
                                 'match_type': match_type}
                                                  )

                            if len(match.groups()) == 1:

                                residual_address = repl_char_from_string(residual_address, match.span(1))

                            elif len(match.groups()) == 2:

                                match_start, match_end = match.span(2)

                                residual_address = repl_char_from_string(residual_address, (match_start, match_end))

                    if len(l_match_results) >= 2:

                        break

                    match_type += 1

                l_match_results = sorted(l_match_results, key=lambda x: span_end(x['match'].span()), reverse=True)

                for i, match_result in enumerate(l_match_results):

                    if i == 0:
                        field_name = 'THOROUGHFARE'
                    elif i == 1:
                        field_name = 'THOROUGHFARE_EXTRA'
                    else:
                        field_name = 'THOROUGHFARE'

                    if len(match_result['match'].groups()) == 1:

                        span[field_name] = match_result['match'].span(1)
                        string[field_name] = match_result['match'].group(1)

                        residual_address = repl_char_from_string(residual_address, match_result['match'].span(1))

                    elif len(match_result['match'].groups()) == 2:

                        span[field_name] = match_result['match'].span(2)
                        string[field_name] = match_result['match'].group(2)

                        residual_address = repl_char_from_string(residual_address, match_result['match'].span(2))

            else:

                match_list = []

                for regex in l_regex:

                    match = re.search(pattern=regex, string=residual_address)

                    if match is not None:
                        match_list.append(match)

                if len(match_list):

                    match = max(match_list, key=lambda x: len(x.group(1)))

                    if key == 'POSTCODE':
                        span[key] = match.span(0)
                        string[key] = ' '.join([match.group(1), match.group(2)])
                        string['PC1'] = match.group(1)
                        string['PC2'] = match.group(2)

                        residual_address = repl_char_from_string(residual_address, match.span(0))
                    else:
                        span[key] = match.span(1)
                        string[key] = match.group(1)

                        residual_address = repl_char_from_string(residual_address, match.span(1))

        string['residual_address'] = residual_address

        return string

    def parse(self, address, method='fast'):

        parsed = {'STRING': {}, 'FOR_QUERY': {}}

        if method == 'fast':

            parsed.update(fast_parsing(address))
            parsed['FOR_QUERY'].update(parsed['NUMBER_LIKE'])
            parsed['FOR_QUERY']['TEXTUAL'] = parsed['TEXTUAL']
            try:
                parsed['FOR_QUERY']['POSTCODE'] = parsed['POSTCODE']
                parsed['FOR_QUERY'].update(parsed['POSTCODE_SPLIT'])
            except KeyError:
                pass

            return parsed

        elif method == 'all':

            parsed.update(fast_parsing(address))
            parsed['FOR_QUERY'].update(parsed['NUMBER_LIKE'])
            parsed['FOR_QUERY']['TEXTUAL'] = parsed['TEXTUAL']
            try:
                parsed['FOR_QUERY']['POSTCODE'] = parsed['POSTCODE']
                parsed['FOR_QUERY'].update(parsed['POSTCODE_SPLIT'])
            except KeyError:
                pass

            string, need_fuzzy_parsing = self.exact_parsing(address)
            parsed['STRING'].update(string)
            parsed['FOR_QUERY'].update(string)

            if need_fuzzy_parsing:
                string = self.fuzzy_parsing(address)
                parsed['STRING'].update(string)
                parsed['FOR_QUERY'].update(string)

        else:
            raise ValueError('`method` must be `fast` or `all`')

        return parsed


NUMBER_LIKE_MASTER_REGEX = re.compile(
    r"(FLAT|UNIT|HOUSE|BUILDING|ROOM|BLOCK|BONDS?|FL|PF|BF|GF|APARTMENT|\(F|F|-|\()? ?"
    r"((\dF\d+)|(\d+[A-Z]\d+)|([A-EG-Z])?(\d+)([A-Z])?|(?<!')(^|\b)([A-Z]|GROUND)($|\b)(?!'))"
    r"\)?")

POSTCODE_MASTER_REGEX = re.compile(r'([A-Z]{1,2})([0-9][A-Z0-9]?)(?: *\n?)?([0-9])([A-Z])([A-Z])')

SPLIT_MASTER_REGEX = re.compile(r"(?<=\d)(?=\D)|(?=\d)(?<=\D)")


def fast_parsing(address):

    p = NUMBER_LIKE_MASTER_REGEX

    split_pattern = SPLIT_MASTER_REGEX

    parsed = {}

    p_pc = POSTCODE_MASTER_REGEX

    match_pc = re.search(p_pc, address)

    string_pc = {}

    if match_pc is not None:

        string_pc['pc_area'], string_pc['pc_district'], string_pc['pc_sector'], string_pc['pc_unit_0'], string_pc['pc_unit_1'] = \
            [match_pc.group(i) for i in range(1, 6)]

        postcode = ' '.join([''.join([string_pc['pc_area'], string_pc['pc_district']]),
                             ''.join([string_pc['pc_sector'], string_pc['pc_unit_0'], string_pc['pc_unit_1']])])

        parsed['POSTCODE_SPLIT'] = string_pc
        parsed['POSTCODE'] = postcode

        address = repl_char_from_string(address, match_pc.span())

    matches = list(re.finditer(p, address))

    prefix = [(match.group(1) is not None) or (re.search(r'\dF\d', match.group(2)) is not None) for match in matches]

    units = [re.split(split_pattern, match.group(2))
             if re.search(r'\dF\d', match.group(2)) is None else [match.group(2)]
             for match in matches]

    units_tmp = units.copy()

    # Fix units

    for i in range(len(matches) - 1):
        if (matches[i+1].span(2)[0] - matches[i].span(2)[1]) == 1:
            char = address[matches[i].span(2)[1]:matches[i+1].span(2)[0]]
            if char != ' ':
                units[i + 1] = units_tmp[i] + units_tmp[i + 1]
                units[i] = []
                units_tmp = units.copy()

                prefix[i] = True
                prefix[i + 1] = True

    for i in range(len(matches) - 1):
        if (not prefix[i]) and prefix[i + 1]:
            p = i + 2
            while (p < len(matches)) and (prefix[p]):
                p += 1
            p -= 1

            units[i], units[p] = units[p], units[i]

    res = flatten(units[::-1])

    if len(res) < 5:
        res += [''] * (5 - len(res))
    elif len(res) > 5:
        res = res[:5]

    for match in matches:
        address = repl_char_from_string(address, match.span())

    string_number_like = {'number_like_%s' % i: res[i] for i in range(len(res))}

    parsed['NUMBER_LIKE'] = string_number_like
    parsed['TEXTUAL'] = address

    return parsed


# import time
#
#
# sql_db = SqlDB('/home/huayu_ssh/PycharmProjects/dres_r/db/scotland_curl')
# parser = RuleParserFast(sql_db)
#
# address = """7 F3 GRANVILLE TERRACE, EDINBURGH"""

# Exact parsing
#
# t0 = time.time()
#
# for i in range(100):
# parsed = parser.parse(address, method='all')
#
# parser.fuzzy_parsing(address)
#
# print(time.time() - t0)
#
# fast_parsing(address)
#
# parser.exact_parsing(address)
# parser.fuzzy_parsing(address)
#

# print(parsed)
# t0 = time.time()
#
# for i in range(100):
#     parsed = parser.parse(address, method='all')
#
# print(time.time() - t0)
# print(parsed)

# r'(%s) ((?:[A-Z]+ ){1,3}(?:%s))(?:\s|$|,|\n)' % ('\d+', '|'.join(parser.l_thoroughfare_names))

#
# # Fussy parsing
#
# t0 = time.time()
#
# for i in range(10):
#     parser.fuzzy_parsing(address)
#
# print(time.time() - t0)
#
#
# t0 = time.time()
#
# for i in range(100):
#
#     for p in parser.regex_rules['POST_TOWN']:
#         re.search(p, address)
#
# print(time.time() - t0)
#
#
# t0 = time.time()
#
# for i in range(100):
#
#     for p in parser.l_posttown_names:
#         p in address
#
# print(time.time() - t0)
#
# t0 = time.time()
#
# for i in range(1000):
#
#     fast_parsing(address)
#
# print(time.time() - t0)
#
#

#
#
# matches = re.finditer(NUMBER_LIKE_MASTER_REGEX, address)
#
# for match in matches:
#     print(match.span())
#     print(match.group())
#     print(match.groups())
#
# parse_number_like(address)
#






import sqlite3
import pandas as pd


class SqlDBInMemory:

    def __init__(self):
        self.columns_dict = {}

    @staticmethod
    def get_conn():
        return sqlite3.connect('file:cachedb?mode=memory&cache=shared', timeout=10)

    def load_csv(self, path, table_name, chunksize=10000):

        print('Start loading DB to memory')
        db_chunks = pd.read_csv(path, chunksize=chunksize, dtype='object', index_col=0)

        i = 0

        while 1:
            try:
                df = next(db_chunks)
                i += 1
                df.fillna('', inplace=True)
                df.to_sql(name=table_name, con=self.get_conn(), if_exists='append', dtype='TEXT', index=False)
                print('Chunk %s loaded' % i, end='\r')

                self.columns_dict[table_name] = df.columns.to_list()

            except StopIteration:
                break

        print('DB Loading Finished')

    def get_columns_of_table(self, table_name):
        return self.columns_dict[table_name]

    def sql_query(self, query):

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)
        res = cur.fetchall()
        cur.close()
        conn.close()
        return res

    def create_index(self, table_name, columns):

        index_name = '__'.join(columns)

        sql = "CREATE INDEX IF NOT EXISTS %s ON %s(%s)" % (
            index_name,
            table_name,
            ', '.join(columns)
        )

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(sql)
        cur.close()
        conn.close()

    def close(self):
        conn = self.get_conn()
        conn.close()
